/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore
National Laboratory.
*/
#include <pbs_config.h>   /* the master config generated by configure */

#include <assert.h>
#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pbs_ifl.h>
#include "list_link.h"
#include "attribute.h"
#include "pbs_error.h"

static char ident[] = "@(#) $RCSfile: attr_fn_size.c,v $ $Revision: 2.0 $";

/*
 * This file contains functions for manipulating attributes of type
 *	size, which is an integer optionally followed by k,K,m,M,g,
 *	G,t, or T, optionally followed by w,W,b,B.
 *	If 'w' or 'W' is not specified, b for bytes is assumed.
 *
 * The attribute has functions for:
 *	Decoding the value string to the machine representation.
 *	Encoding the internal attribute to external form
 *	Setting the value by =, + or - operators.
 *	Comparing a (decoded) value with the attribute value.
 *
 * Some or all of the functions for an attribute type may be shared with
 * other attribute types.
 * 
 * The prototypes are declared in "attribute.h"
 *
 * --------------------------------------------------
 * The Set of Attribute Functions for attributes with
 * value type "size"
 * --------------------------------------------------
 */


/*
 * decode_size - decode size into attribute structure
 *
 *	Returns: 0 if ok
 *		>0 error number if error
 *		*patr elements set
 */

int decode_size(patr, name, rescn, val)
	struct attribute *patr;
	char *name;		/* attribute name */
	char *rescn;		/* resource name, unused here */
	char *val;		/* attribute value */
{
	int to_size A_((char *, struct size_value *));

	patr->at_val.at_size.atsv_num   = 0;
	patr->at_val.at_size.atsv_shift = 0;
	if ((val != (char *)0) && (strlen(val) != 0)) {
		if (to_size(val, &patr->at_val.at_size) != 0)
			return (PBSE_BADATVAL);
		patr->at_flags |= ATR_VFLAG_SET | ATR_VFLAG_MODIFY;
	} else {
		patr->at_flags = (patr->at_flags & ~ATR_VFLAG_SET) |
				 ATR_VFLAG_MODIFY;
	}
	return (0);
}

/*
 * encode_size - encode attribute of type size into attr_extern
 *
 *	Returns: >0 if ok
 *		 =0 if no value, no attrlist link added
 *		 <0 if error
 */
/*ARGSUSED*/

#define CVNBUFSZ 23

int encode_size(attr, phead, atname, rsname, mode)
	attribute	*attr;	  /* ptr to attribute */
	list_head	*phead;	  /* head of attrlist list */
	char		*atname;  /* attribute name */
	char		*rsname;  /* resource name or null */
	int		mode;	  /* encode mode, unused here */
{
	size_t	     ct;
	char	     cvnbuf[CVNBUFSZ];
	svrattrl *pal;
	void from_size A_((struct size_value *, char *));

	if ( !attr )
		return (-1);
	if ( !(attr->at_flags & ATR_VFLAG_SET))
		return (0);

	from_size(&attr->at_val.at_size, cvnbuf);
	ct = strlen(cvnbuf) + 1;

	pal = attrlist_create(atname, rsname, ct);
	if (pal == (svrattrl *)0)
		return (-1);

	(void)memcpy(pal->al_value, cvnbuf, ct);
	pal->al_flags = attr->at_flags;
	append_link(phead, &pal->al_link, pal);
	
	return (1);
}

/*
 * set_size - set attribute A to attribute B,
 *	either A=B, A += B, or A -= B
 *
 *	Returns: 0 if ok
 *		>0 if error
 */

int set_size(attr, new, op)
	struct attribute *attr;
	struct attribute *new;
	enum batch_op op;
{
	unsigned long old;
	struct size_value tmpa;	/* the two temps are used to insure that the */
	struct size_value tmpn;	/* real attributes are not changed if error  */
	int normalize_size A_((struct size_value *a, struct size_value *b,
			      struct size_value *c, struct size_value *d));

	assert ( attr && new && (new->at_flags & ATR_VFLAG_SET) );

	switch (op) {
	    case SET:	attr->at_val.at_size.atsv_num   = new->at_val.at_size.atsv_num;
	    		attr->at_val.at_size.atsv_shift = new->at_val.at_size.atsv_shift;
			attr->at_val.at_size.atsv_units = new->at_val.at_size.atsv_units;
			break;

	    case INCR:	if (normalize_size(&attr->at_val.at_size, 
			    &new->at_val.at_size, &tmpa, &tmpn) < 0)
				return (PBSE_BADATVAL);
			old  = tmpa.atsv_num;
			tmpa.atsv_num += tmpn.atsv_num;
			if (tmpa.atsv_num < old)
				return (PBSE_BADATVAL);
			attr->at_val.at_size = tmpa;
			break;

	    case DECR:	if (normalize_size(&attr->at_val.at_size, 
			    &new->at_val.at_size, &tmpa, &tmpn) < 0)
				return (PBSE_BADATVAL);
			old  = tmpa.atsv_num;
			tmpa.atsv_num -= tmpn.atsv_num;
			if (tmpa.atsv_num > old)
				return (PBSE_BADATVAL);
			attr->at_val.at_size = tmpa;
			break;

	    default:	return (PBSE_INTERNAL);
	}
	attr->at_flags |= ATR_VFLAG_SET | ATR_VFLAG_MODIFY;

	return (0);
}

/*
 * comp_size - compare two attributes of type size
 *
 *	Returns: +1 if 1st > 2nd
 *		  0 if 1st == 2nd
 *		 -1 if 1st < 2nd
 */

int comp_size(attr, with)
	struct attribute *attr;
	struct attribute *with;
{
	struct size_value tmpa;
	struct size_value tmpw;
	int normalize_size A_((struct size_value *a, struct size_value *b,
			       struct size_value *c, struct size_value *d));

	if (normalize_size(&attr->at_val.at_size, &with->at_val.at_size,
            &tmpa, &tmpw) != 0) {
		if (tmpa.atsv_shift > 
		    tmpw.atsv_shift)
			return (1);
		else if (tmpa.atsv_shift < 
		    tmpw.atsv_shift)
			return (-1);
		else 
			return (0);
	} else if (tmpa.atsv_num > tmpw.atsv_num)
		return (1);
	else if (tmpa.atsv_num < tmpw.atsv_num)
		return (-1);
	else
		return (0);

}

/*
 * free_size - use free_null to (not) free space
 */

/*
 * normalize_size - normalize two size values, adjust so the shift
 *	counts are the same.
 */

int normalize_size (a, b, ta, tb)
	struct size_value *a;
	struct size_value *b;
	struct size_value *ta;
	struct size_value *tb;
{
	int adj;
	unsigned long temp;

	/*
	 * we do the work in copies of the original attributes
	 * to preserve the original (in case of error)
	 */
	*ta = *a;
	*tb = *b;

	/* if either unit is in bytes, then both must be */

	if ((ta->atsv_units == ATR_SV_WORDSZ) &&
	    (tb->atsv_units != ATR_SV_WORDSZ)) {
		ta->atsv_num *= sizeof(int);
		ta->atsv_units = ATR_SV_BYTESZ;
	} else if ((ta->atsv_units != ATR_SV_WORDSZ) &&
		   (tb->atsv_units == ATR_SV_WORDSZ)) {
		tb->atsv_num *= sizeof(int);
		tb->atsv_units = ATR_SV_BYTESZ;
	}
		
	adj = ta->atsv_shift - tb->atsv_shift;

	if (adj > 0) {
		temp = ta->atsv_num;
		if ((adj > sizeof (int) * 8) ||
		    (((temp << adj) >> adj) != ta->atsv_num))
			return (-1);	/* would overflow */
		ta->atsv_shift = tb->atsv_shift;
		ta->atsv_num   = ta->atsv_num << adj;
	} else if (adj < 0) {
		adj = -adj;
		temp = tb->atsv_num;
		if ((adj > sizeof (int) * 8) ||
		    (((temp << adj) >> adj) != tb->atsv_num))
			return (-1);	/* would overflow */
		tb->atsv_shift = ta->atsv_shift;
		tb->atsv_num   = tb->atsv_num << adj;
	}
	return (0);
}


/* 
 * to_size - decode the value string TO a size_value structure
 *
 *	This function decodes a value string into a size_value struct.
 */

int to_size(val, psize)
	char	 *val;
	struct size_value *psize;
{
	int   havebw = 0;
	char *pc;

	psize->atsv_units = ATR_SV_BYTESZ;
	psize->atsv_num = strtol(val, &pc, 0);
	if (pc == val)		/* no numeric part */
		return (PBSE_BADATVAL);

	switch (*pc) {
	    case '\0':	break;
	    case 'k':
	    case 'K':	psize->atsv_shift = 10;
			break;
	    case 'm':
	    case 'M':	psize->atsv_shift = 20;
			break;
	    case 'g':
	    case 'G':	psize->atsv_shift = 30;
			break;
	    case 't':
	    case 'T':	psize->atsv_shift = 40;
			break;
	    case 'p':
	    case 'P':	psize->atsv_shift = 50;
			break;
	    case 'b':
	    case 'B':	havebw = 1;
			break;
	    case 'w':
	    case 'W':	havebw = 1;
			psize->atsv_units = ATR_SV_WORDSZ;
			break;
	    default:	return (PBSE_BADATVAL);	 /* invalid string */
	}
	if (*pc != '\0') 
		pc++;
	if (*pc != '\0') {
		if (havebw)
			return (PBSE_BADATVAL);  /* invalid string */
		switch (*pc) {
		    case 'b':
		    case 'B':	break;
		    case 'w':
		    case 'W':	psize->atsv_units = ATR_SV_WORDSZ;
				break;
		    default:	return (PBSE_BADATVAL);
		}
	}
	return (0);
}

/*
 * from_size - encode a string FROM a size_value structure
 */

void from_size(psize, cvnbuf)
	struct size_value *psize;
	char		  *cvnbuf;
{

	(void)sprintf(cvnbuf, "%lu", psize->atsv_num);
	switch (psize->atsv_shift) {
		case  0:	break;
		case 10:	strcat(cvnbuf, "k");
				break;
		case 20:	strcat(cvnbuf, "m");
				break;
		case 30:	strcat(cvnbuf, "g");
				break;
		case 40:	strcat(cvnbuf, "t");
				break;
		case 50:	strcat(cvnbuf, "p");
	}
	if (psize->atsv_units & ATR_SV_WORDSZ)
		strcat(cvnbuf, "w");
	else
		strcat(cvnbuf, "b");
}
