/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore
National Laboratory.
*/
#include <pbs_config.h>   /* the master config generated by configure */

#include <assert.h>
#include <ctype.h>
#include <memory.h>
#ifndef NDEBUG
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include "pbs_ifl.h"
#include "list_link.h"
#include "attribute.h"
#include "pbs_error.h"

static char ident[] = "@(#) $RCSfile: attr_fn_str.c,v $ $Revision: 2.0 $";

/*
 * This file contains functions for manipulating attributes of type string
 *
 * Then there are a set of functions for each type of attribute:
 *	string
 *
 * Each set has functions for:
 *	Decoding the value string to the internal representation.
 *	Encoding the internal attribute form to external form
 *	Setting the value by =, + or - operators.
 *	Comparing a (decoded) value with the attribute value.
 *
 * Some or all of the functions for an attribute type may be shared with
 * other attribute types.
 * 
 * The prototypes are declared in "attribute.h"
 *
 * -------------------------------------------------
 * Set of general attribute functions for attributes
 * with value type "string"
 * -------------------------------------------------
 */

/*
 * decode_str - decode string into string attribute
 *
 *	Returns: 0 if 0k
 *		>0 error number if error
 *		*patr members set
 */

int decode_str(patr, name, rescn, val)
	struct attribute *patr;
	char *name;		/* attribute name */
	char *rescn;		/* resource name - unused here */
	char *val;		/* attribute value */
{
	size_t len;

	if ((patr->at_flags & ATR_VFLAG_SET) && (patr->at_val.at_str))
		(void)free(patr->at_val.at_str);

	if ((val != (char *)0) && ((len = strlen(val) + 1) > 1)) {
		patr->at_val.at_str = malloc((unsigned) len);
		if (patr->at_val.at_str == (char *)0)
			return (PBSE_SYSTEM);
		(void)strcpy(patr->at_val.at_str, val);
		patr->at_flags |= ATR_VFLAG_SET | ATR_VFLAG_MODIFY;
	} else {
		patr->at_flags = (patr->at_flags & ~ATR_VFLAG_SET) |
				 ATR_VFLAG_MODIFY;
		patr->at_val.at_str = (char *)0;
	}
	return (0);
}

/*
 * encode_str - encode attribute of type ATR_TYPE_STR into attr_extern
 *
 *	Returns: >0 if ok
 *		 =0 if no value, so no link added to list
 *		 <0 if error
 */
/*ARGSUSED*/

int encode_str(attr, phead, atname, rsname, mode)
	attribute	*attr;	  /* ptr to attribute */
	list_head	*phead;	  /* head of attrlist */
	char		*atname;  /* name of attribute */
	char		*rsname;  /* resource name or null */
	int		 mode;	  /* encode mode, unused here */
	
{
	svrattrl *pal;

	if ( !attr )
		return (-1);
	if ( !(attr->at_flags & ATR_VFLAG_SET) || !attr->at_val.at_str ||
	    (*attr->at_val.at_str == '\0'))
		return (0);

	pal = attrlist_create(atname,rsname,(int)strlen(attr->at_val.at_str)+1);
	if (pal == (svrattrl *)0)
		return (-1);

	(void)strcpy(pal->al_value, attr->at_val.at_str);
	pal->al_flags = attr->at_flags;
	append_link(phead, &pal->al_link, pal);

	return (1);
}

/*
 * set_str - set attribute value based upon another
 *
 *	A+B --> B is concatenated to end of A
 *	A=B --> A is replaced with B
 *	A-B --> If B is a substring at the end of A, it is stripped off
 *
 *	Returns: 0 if ok
 *		>0 if error
 */

int set_str(attr, new, op)
	struct attribute *attr;
	struct attribute *new;
	enum batch_op op;
{
	char	*new_value;
	char	*p;
	size_t nsize;

	assert( attr && new && new->at_val.at_str && (new->at_flags & ATR_VFLAG_SET));
	nsize = strlen(new->at_val.at_str) + 1;	/* length of new string */
	if ((op == INCR) && !attr->at_val.at_str)
		op = SET;	/* no current string, change INCR to SET */

	switch (op) {

	    case SET:	/* set is replace old string with new */

		if (attr->at_val.at_str)
			(void)free(attr->at_val.at_str);
		if ((attr->at_val.at_str = malloc(nsize)) == (char *)0)
			return (PBSE_SYSTEM);
		(void)strcpy(attr->at_val.at_str, new->at_val.at_str);
		break;

	    case INCR:	/* INCR is concatenate new to old string */

		nsize += strlen(attr->at_val.at_str);
		if (attr->at_val.at_str)
			new_value = realloc(attr->at_val.at_str, nsize);
		else
			new_value = malloc(nsize);
		if (new_value == (char *)0)
			return (PBSE_SYSTEM);
		attr->at_val.at_str = new_value;
		(void)strcat(attr->at_val.at_str, new->at_val.at_str);
		break;

	    case DECR:	/* DECR is remove substring if match, start at end */

		if (!attr->at_val.at_str)
			break;

		if (--nsize == 0)
			break;
		p = attr->at_val.at_str + strlen(attr->at_val.at_str) - nsize;
		while (p >= attr->at_val.at_str) {
			if (strncmp(p, new->at_val.at_str, (int)nsize) == 0) {
				do {
					*p = *(p + nsize);
				} while (*p++);
		  	}
			p--;
		}
		break;

	    default:	return (PBSE_INTERNAL);
	}
	if ((attr->at_val.at_str != (char *)0) && (*attr->at_val.at_str !='\0'))
		attr->at_flags |= ATR_VFLAG_SET | ATR_VFLAG_MODIFY;
	else
		attr->at_flags &= ~ATR_VFLAG_SET;

	return (0);
}

/* 
 * comp_str - compare two attributes of type ATR_TYPE_STR
 *
 *	Returns: see strcmp(3)
 */

int comp_str(attr, with)
	struct attribute *attr;
	struct attribute *with;
{
	if ( !attr || !attr->at_val.at_str)
		return (-1);
	return (strcmp(attr->at_val.at_str, with->at_val.at_str));
}

/*
 * free_str - free space malloc-ed for string attribute value
 */
void free_str(attr)
	struct attribute *attr;
{
	if ((attr->at_flags & ATR_VFLAG_SET) && (attr->at_val.at_str)) {
		(void)free(attr->at_val.at_str);
	}
	attr->at_val.at_str = (char *)0;
	attr->at_flags &= ~ATR_VFLAG_SET;
}
