/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore
National Laboratory.
*/
#include <pbs_config.h>   /* the master config generated by configure */

#include <assert.h>
#include <ctype.h>
#include <memory.h>
#ifndef NDEBUG
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <pbs_ifl.h>
#include "list_link.h"
#include "attribute.h"
#include "resource.h"
#include "pbs_error.h"

static char ident[] = "@(#) $RCSfile: attr_fn_unkn.c,v $ $Revision: 2.0 $";

/*
 * This file contains functions for manipulating attributes of an
 * unknown (unrecognized) name (and therefore unknown type).
 * It is a collection point for all "other" attributes, other than
 * the types with specific definition and meaning.
 *
 * Because the type is unknown, it cannot be decoded into a native
 * form.  Thus the attribute is maintained in the attrlist form.
 * Any attribute/value located here will be sent to the Scheduler and it
 * within its rules may do as it choses with them.
 *
 * The prototypes are declared in "attribute.h"
 *
 * ----------------------------------------------------------------------------
 * Attribute functions for attributes with value type "unknown"
 * ----------------------------------------------------------------------------
 */

/* External Global Items */


/* private functions */


/*
 * decode_unkn - decode a pair of strings (name and value) into the Unknown
 *	type attribute/resource which is maintained as a "svrattrl", a
 *	linked list of structures containing strings.
 *
 *	Returns: 0 if ok,
 *		>0 error number if error,
 *		*patr members set
 */

int decode_unkn(patr, name, rescn, value)
	struct attribute *patr;		/* May be Modified on Return */
	char *name;
	char *rescn;
	char *value;
{
	svrattrl *entry;
	size_t	     valln;


	if (patr == (attribute *)0)
		return (PBSE_INTERNAL);

	if ( !(patr->at_flags & ATR_VFLAG_SET))
		CLEAR_HEAD(patr->at_val.at_list);

	if (name == (char *)0)
		return (PBSE_INTERNAL);

	if (value == (char *)0)
		valln = 0;
	else
		valln = strlen(value) + 1;

	entry = attrlist_create(name, rescn, valln);
	if (entry == (svrattrl *)0)
		return (PBSE_SYSTEM);

	if (valln)
	    memcpy(entry->al_value, value, valln);
	
	append_link(&patr->at_val.at_list, &entry->al_link, entry);
	patr->at_flags |= ATR_VFLAG_SET | ATR_VFLAG_MODIFY;
	return (0);
}
	

/*
 * encode_unkn - encode attr of unknown type into attrlist form
 *
 * Here things are different from the typical attribute.  Most have a
 * single value to be encoded.  But "the unknown" attribute may have a whole
 * list.
 *
 * This function does not use the parent attribute name, after all "_other_"
 * is rather meaningless.  In addition, each unknown already is in an 
 * attrlist form.
 *
 * Thus for each entry in the list, encode_unkn duplicates the existing
 * attrlist struct and links the copy into the list.
 *
 *	Returns:>=0 if ok, the total encoded size of all resources.
 *		-1 if encoded value would not fit into buffer
 *		-2 if error
 */
/*ARGSUSED*/

int encode_unkn(attr, phead, atname, rsname, mode)
	attribute	*attr;	  /* ptr to attribute to encode */
	list_head	*phead;	  /* list to place entry in */
	char 		*atname;  /* attribute name, not used here */
	char	 	*rsname;  /* resource name, not used here */
	int		 mode;	  /* encode mode, unused here */
{
	svrattrl *plist;
	svrattrl *pnew;

	if ( !attr )
		return (-2);

	plist= (svrattrl *)GET_NEXT(attr->at_val.at_list);
	if (plist == (svrattrl *)0)
		return (0);

	while ( plist != (svrattrl *)0) {
		pnew = (svrattrl *)malloc(plist->al_tsize);
		if (pnew == (svrattrl *)0)
			return (-1);
		CLEAR_LINK(pnew->al_link);
		pnew->al_tsize = plist->al_tsize;
		pnew->al_nameln = plist->al_nameln;
		pnew->al_rescln = plist->al_rescln;
		pnew->al_valln  = plist->al_valln;
		pnew->al_flags   = plist->al_flags;

		pnew->al_name  = (char *)pnew + sizeof (svrattrl);
		(void)memcpy(pnew->al_name, plist->al_name, plist->al_nameln);
		if (plist->al_rescln) {
			pnew->al_resc = pnew->al_name + pnew->al_nameln;
			(void)memcpy(pnew->al_resc, plist->al_resc,
				     plist->al_rescln);
		} else {
			pnew->al_resc = (char *)0;
		}
		if (plist->al_valln) {
			pnew->al_value = pnew->al_name + pnew->al_nameln + 
					 pnew->al_rescln;
			(void)memcpy(pnew->al_value, plist->al_value,
				     pnew->al_valln);
		}
		append_link(phead, &pnew->al_link, pnew);

		plist = (svrattrl *)GET_NEXT(plist->al_link);
	}
	return (1);
}
	


/* 
 * set_unkn - set value of attribute of unknown type  to another
 *
 *	Each entry in the list headed by the "new" attribute is appended
 *	to the list headed by "old".
 *
 *	All operations, set, incr, and decr, map to append.
 *	Returns: 0 if ok
 *		>0 if error
 */

/*ARGSUSED*/
int set_unkn(old, new, op)
	struct attribute *old;
	struct attribute *new;
	enum batch_op op;
{
	svrattrl *plist;
	svrattrl *pnext;
	
	assert(old && new && (new->at_flags & ATR_VFLAG_SET));

	plist = (svrattrl *)GET_NEXT(new->at_val.at_list);
	while ( plist != (svrattrl *)0) {
		pnext = (svrattrl *)GET_NEXT(plist->al_link);
		delete_link(&plist->al_link);
		append_link(&old->at_val.at_list, &plist->al_link, plist);
		plist = pnext;
	}
	old->at_flags |= ATR_VFLAG_SET | ATR_VFLAG_MODIFY;
	return (0);
}

/*
 * comp_unkn - compare two attributes of type ATR_TYPE_RESR
 *
 *	How do you compare something when you don't know what it is...
 *	So, always returns +1
 *
 */

int comp_unkn(attr, with)
	struct attribute *attr;
	struct attribute *with;
{
	return (1);
}
	
/*
 * free_unkn - free space associated with attribute value
 *
 *	For each entry in the list, it is delinked, and freed.
 */

void free_unkn(pattr)
	attribute *pattr;
{
	svrattrl *plist;

	if (pattr->at_flags & ATR_VFLAG_SET) {
		while ((plist = (svrattrl *)GET_NEXT(pattr->at_val.at_list)) !=
								(svrattrl *)0) {
			delete_link(&plist->al_link);
			(void)free(plist);
		}
	}
	CLEAR_HEAD(pattr->at_val.at_list);
	pattr->at_flags &= ~ATR_VFLAG_SET;
}
