/*
 * Synopsis:
 *	unsigned short disrus(int stream, int *retval)
 *
 *	Gets a Data-is-Strings unsigned integer from <stream>, converts it into
 *	an unsigned short, and returns it.  The unsigned integer in <stream>
 *	consists of a counted string of digits, starting with a zero, which
 *	represents the number.  If the number doesn't lie between 0 and 9,
 *	inclusive, it is preceeded by at least one count.
 *
 *	This format for character strings representing unsigned integers can
 *	best be understood through the decoding algorithm:
 *
 *	1. Initialize the digit count to 1.
 *
 *	2. Read the next digit; if it is a plus sign, go to step (4); if it
 *	   is a minus sign, post an error.
 *
 *	3. Decode a new count from the digit decoded in step (2) and the next
 *	   count - 1 digits; repeat step (2).
 *
 *	4. Decode the next count digits as the unsigned integer.
 *
 *	*<retval> gets DIS_SUCCESS if everything works well.  It gets an error
 *	code otherwise.  In case of an error, the <stream> character pointer is
 *	reset, making it possible to retry with some other conversion strategy.
 */

#include <pbs_config.h>   /* the master config generated by configure */

#include <assert.h>
#include <stddef.h>

#include "dis.h"
#include "dis_.h"
#undef disrus

unsigned short disrus(stream, retval)
    int			stream;
    int			*retval;
    {
	int		locret;
	int		negate;
	unsigned	value;

	assert(retval != NULL);
	assert(disr_commit != NULL);

	locret = disrsi_(stream, &negate, &value, 1);
	if (locret != DIS_SUCCESS) {
		value = 0;
	} else if (negate) {
		value = 0;
		locret = DIS_BADSIGN;
	} else if (value > USHRT_MAX) {
		value = USHRT_MAX;
		locret = DIS_OVERFLOW;
	}
	*retval = ((*disr_commit)(stream, locret == DIS_SUCCESS) < 0) ?
	    DIS_NOCOMMIT : locret;
	return ((unsigned short)value);
}
