/*
 * Synopsis:
 * 	int diswl_(int stream, dis_long_double_t value, unsigned int ndigs)
 *
 *	Converts <value> into a Data-is-Strings floating point number and sends
 *	it to <stream>.  The converted number consists of two consecutive signed
 *	integers.  The first is the coefficient, at most <ndigs> long, with its
 *	implied decimal point at the low-order end.  The second is the exponent
 *	as a power of 10.
 *
 *	This function is only invoked through the macros, diswf, diswd, and
 *	diswl, which are defined in the header file, dis.h.
 *
 *	Returns DIS_SUCCESS if everything works well.  Returns an error code
 *	otherwise.  In case of an error, no characters are sent to <stream>.
 */

#include <pbs_config.h>   /* the master config generated by configure */

#include <assert.h>
#include <stddef.h>
#include <stdio.h>

#include "dis.h"
#include "dis_.h"

int diswl_(stream, value, ndigs)
    int			stream;
    dis_long_double_t		value;
    unsigned		ndigs;
    {
	int		c;
	int		expon;
	int		negate;
	int		retval;
	unsigned	pow2;
	char		*cp;
	char		*ocp;
	dis_long_double_t	ldval;

	assert(ndigs > 0 && ndigs <= LDBL_DIG);
	assert(stream >= 0);
	assert(dis_puts != NULL);
	assert(disw_commit != NULL);

/* Make zero a special case.  If we don't it will blow exponent		*/
/* calculation.								*/
	if (value == 0.0L) {
		retval = (*dis_puts)(stream, "+0+0", 4) < 0 ?
		    DIS_PROTO : DIS_SUCCESS;
		return (((*disw_commit)(stream, retval == DIS_SUCCESS) < 0) ?
		       DIS_NOCOMMIT : retval);
	}
/* Extract the sign from the coefficient.				*/
        ldval = (negate = value < 0.0L) ? -value : value;
/* Detect and complain about the infinite form.				*/
	if (ldval > LDBL_MAX)
	        return (DIS_HUGEVAL);
/* Compute the integer part of the log to the base 10 of ldval.  As a	*/
/* byproduct, reduce the range of ldval to the half-open interval,      */
/* [1, 10).								*/
	if (dis_lmx10 == 0)
	        disi10l_();
	expon = 0;
	pow2 = dis_lmx10 + 1;
	if (ldval < 1.0L) {
		do {
			if (ldval < dis_ln10[--pow2]) {
				ldval *= dis_lp10[pow2];
				expon += 1 << pow2;
			}
		} while (pow2);
		ldval *= 10.0;
		expon = -expon - 1;
	} else {
		do {
			if (ldval >= dis_lp10[--pow2]) {
				ldval *= dis_ln10[pow2];
				expon += 1 << pow2;
			}
		} while (pow2);
	}
/* Round the value to the last digit					*/
	ldval += 5.0L * disp10l_(-ndigs);
	if (ldval >= 10.0L) {
		expon++;
		ldval *= 0.1L;
	}
/* Starting in the middle of the buffer, convert coefficient digits,	*/
/* most significant first.						*/
	ocp = cp = &dis_buffer[DIS_BUFSIZ - ndigs];
	do {
		c = ldval;
		ldval = (ldval - c) * 10.0L;
		*ocp++ = c + '0';
	} while (--ndigs);
/* Eliminate trailing zeros.						*/
	while (*--ocp == '0');
/* The decimal point is at the low order end of the coefficient		*/
/* integer, so adjust the exponent for the number of digits in the	*/
/* coefficient.								*/
	ndigs = ++ocp - cp;
	expon -= ndigs - 1;
/* Put the coefficient sign into the buffer, left of the coefficient.	*/
	*--cp = negate ? '-' : '+';
/* Insert the necessary number of counts on the left.			*/
	while (ndigs > 1)
	        cp = discui_(cp, ndigs, &ndigs);
/* The complete coefficient integer is done.  Put it out.		*/
	retval = (*dis_puts)(stream, cp, (size_t)(ocp - cp)) < 0 ?
	    DIS_PROTO : DIS_SUCCESS;
/* If that worked, follow with the exponent, commit, and return.	*/
	if (retval == DIS_SUCCESS)
		return (diswsi(stream, expon));
/* If coefficient didn't work, negative commit and return the error.	*/
	return (((*disw_commit)(stream, FALSE) < 0)  ? DIS_NOCOMMIT : retval);
}
