/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore
National Laboratory.
*/
#include <pbs_config.h>   /* the master config generated by configure */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include "portability.h"
#include "log.h"

static char ident[] = "@(#) $RCSfile: chk_file_sec.c,v $ $Revision: 2.0 $";

#ifndef	S_ISLNK
#define	S_ISLNK(m)	(((m) & S_IFMT) == S_IFLNK)
#endif

/*
 * chk_file_sec() - Check file/directory security
 *      Part of the PBS System Security "Feature"
 *
 *	To be secure, all directories (and final file) in path must be:
 *		owned by uid < 10
 *		owned by group < 10 if group writable
 *		not have world writable unless stick bit set & this is allowed.
 *
 * Returns 0 if ok
 *      errno value if not ok, including:
 *              EPERM if not owned by root
 *              ENOTDIR if not file/directory as specified
 *              EACCESS if permissions are not ok
 */

int chk_file_sec(path, isdir, sticky, disallow, fullpath)
	char *path;		/* path to check */
	int   isdir;		/* 1 = path is directory, 0 = file */
	int   sticky;		/* allow write on directory if sticky set */
	int   disallow;		/* perm bits to disallow */
	int   fullpath;		/* recursively check full path */
{
	int    i;
	char  *error_buf;
	char  *pc;
	int    rc = 0;
	struct stat sbuf;
	char   shorter[_POSIX_PATH_MAX];
	char   symlink[_POSIX_PATH_MAX];
	
	if ((*path == '/') && fullpath) {

	    /* check full path starting at root */

	    (void)strcpy(shorter, path);
	    pc  = strrchr(shorter, (int)'/');
	    if ( (pc != (char *)0) && (pc != shorter) ) {

		/*
		 * push "dirname" onto stack, stack will pop back from
		 * root to the given file/directory
 		 */

		*pc = '\0';
		if (rc = chk_file_sec(shorter,1,sticky,S_IWGRP|S_IWOTH,1)) {
			return (rc);
		}
	    }
	}

	if (lstat(path, &sbuf) == -1) {
		rc = errno;
		goto chkerr;
	}

	if (S_ISLNK(sbuf.st_mode) != 0) {
		i = readlink(path, symlink, _POSIX_PATH_MAX);
		if (i < 0) {
			rc = errno;
			goto chkerr;
		}
		*(symlink+i) = '\0';
		if (*symlink == '/') {
			return (chk_file_sec(symlink,isdir,sticky,disallow,fullpath));
		} else {
			(void)strcat(shorter, "/");
			(void)strcat(shorter, symlink);
			return (chk_file_sec(shorter,isdir,sticky,disallow,fullpath));
		}
	}
			
	i = sbuf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO);
	if (sbuf.st_uid > 10) {
		rc = EPERM;

	} else if ( ((isdir == 1) && (S_ISDIR(sbuf.st_mode)==0)) ||
	     ((isdir == 0) && (S_ISREG(sbuf.st_mode)==0)) ) {
		rc = ENOTDIR;

	} else if (i & disallow)	{
		/* if group write, gid must be less than 10 */
		if ( (i & disallow & S_IWGRP) && (sbuf.st_gid > 9) )
			rc = EPERM;
		/* if world write, sticky bit must be set and "sticky" ok */
		if (i & disallow & S_IWOTH) {
			if ((S_ISDIR(sbuf.st_mode) == 0) || 
			    (((sbuf.st_mode & S_ISVTX) == 0) || (sticky != 1)) )
				rc =  EACCES;
		}
	}

chkerr:
	if (rc != 0) {
		if ((error_buf = malloc(1024)) == 0) {
			log_err(rc, "chk_file_sec", "Malloc failed");		
		} else {
			(void)sprintf(error_buf,
					"Security violation with \"%s\"",
					path);
			log_err(rc, "chk_file_sec", error_buf);
			(void)free(error_buf);
		}
	}

	return (rc);
}
