/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National 
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore   
National Laboratory.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <strings.h>
#include <pbs_ifl.h>
#include <log.h>
#include "config.h"
#include "constant.h"
#include "misc.h"
#include "globals.h"
#include "fairshare.h"

static char *ident = "$Id: misc.c,v 2.0 1998/10/14 23:16:52 hender Exp $";

/*
 *
 *	string_dup - duplicate a string
 *
 *	  str - string to duplicate
 *
 *	returns newly allocated string
 *
 */

char *string_dup( char *str )
{
  char *newstr;
  if( ( newstr = (char *) malloc( strlen(str) + 1 ) ) == NULL )
    return NULL;
  
  strcpy(newstr, str);

  return newstr;
}

/*
 *
 *      res_to_num - convert a resource string to a  sch_resource_t to
 *                      kilobytes
 *                      example: 1mb -> 1024 
 *				 1mw -> 1024 * SIZE_OF_WORD
 *
 *      res_str - the resource string
 *
 *      return a number in kilobytes or UNSPECIFIED on error
 *
 */

sch_resource_t res_to_num( char * res_str )
{
  sch_resource_t count = UNSPECIFIED;	/* convert string resource to numeric */
  char *endp;				/* used for strtol() */
  char *endp2;				/* used for strtol() */
  long multiplier;			/* multiplier to count */

  count = strtol( res_str, &endp, 10 );

  if( *endp == ':' )		/* time resource -> convert to seconds */
  {
    count *= 3600;					/* hours */
    if( *endp == ':' )
    {
      count += (strtol( endp+1, &endp2, 10) * 60);	/* minutes */
      if( *endp2 == ':' )
        count += strtol( endp2 + 1, &endp, 10);		/* seconds */
    }
    multiplier = 1;
  }
  else if( *endp == 'k' )
    multiplier = 1;
  else if( *endp == 'm' )
    multiplier = MEGATOKILO;
  else if( *endp == 'g' )
    multiplier = GIGATOKILO;
  else if( *endp == 't' )
    multiplier = TERATOKILO;
  else if( *endp == 'b' )
  {
    count /= KILO;
    multiplier = 1;
  }
  else if( *endp == 'w' )
  {
    count /= KILO;
    multiplier = SIZE_OF_WORD;
  }
  else	/* error */
    multiplier = 1;

  if( *endp != '\0' && *(endp + 1) == 'w' )
    multiplier *= SIZE_OF_WORD;

  return count * multiplier;
}

/*
 *
 *      skip_line - find if the line of the config file needs to be skipped
 *                  due to it being a comment or other means
 *
 *        line - the line from the config file
 *
 *      returns true if the line should be skipped or false if it should be
 *              parsed
 *
 */
int skip_line( char *line )
{
  int skip = 0;				/* whether or not to skil the line */
 
  if( line != NULL )
  {
    while( isspace( (int) *line ) )
      line++;
 
    /* '#' is comment in config files and '*' is comment in holidays file */
    if( line[0] == '\0' || line[0] == '#' || line[0] == '*' )
      skip = 1;
  }
 
  return skip;
}

/*
 *
 *	log - write a log entry to the scheduler log file using log_record
 *
 *	  event - the event type
 *	  class - the event class
 *	  name  - the name of the object
 *	  text  - the text of the message 
 *
 *	returns nothing
 *
 */

void log( int event, int class, char *name, char *text)
{
  if( !(conf.log_filter & event) && text[0] != '\0' )
    log_record(event, class, name, text);
}

/*
 *
 *	break_comma_list - break apart a comma delemetd string into an arary
 *			   of strings
 *
 *	  list - the comma list
 *
 *	returns an array of strings
 *
 */
char **break_comma_list( char *list )
{
  int num_words = 1;			/* number of words delimited by commas*/
  char **arr = NULL;			/* the array of words */
  char *tok;				/* used with strtok() */
  int i;

  if( list != NULL )
  {
    for( i = 0; list[i] != '\0'; i++ )
      if( list[i] == ',' )
	num_words++;
    
    if( ( arr = (char **) malloc( sizeof(char*)  * (num_words + 1) ) ) == NULL )
    {
      perror("Memory Allocation Error");
      return NULL;
    }

    tok = strtok(list, ",");

    for( i = 0; tok != NULL; i++ )
    {
      while( isspace( (int) *tok) )
	tok++;
      
      arr[i] = string_dup(tok);
      tok = strtok(NULL, ",");
    }
    arr[i] = NULL;
  }
  return arr;
}

/* 
 *
 *	free_string_array - free an array of strings with a NULL as a sentinal
 *
 *	  arr - the array to free
 *
 *	returns nothing
 *
 */
void free_string_array( char **arr )
{
  int i;

  if( arr != NULL )
  {
    for( i = 0; arr[i] != NULL; i++)
      free(arr[i]);
    
    free(arr);
  }
}
