#############################################################################
##                           PBS CU-scheduler
##
## CU-sched-body.tcl: scheduler iteration body
##
## (c) 1998  Dirk Grunwald and Artur Klauser, 
##           University of Colorado at Boulder,
##           [grunwald,klauser]@cs.colorado.edu
##
## $Id: CU-sched-body.tcl,v 2.0 1998/10/14 23:19:31 hender Exp $
## $Log: CU-sched-body.tcl,v $
## Revision 2.0  1998/10/14 23:19:31  hender
## PBS Release 2.0
##
## Revision 1.1  1998/08/06 21:20:10  hender
## PR #1234 - Dec Alpha mods and scheduler changes from U Colorado
##
## Revision 1.1  1998/08/05 22:20:57  grunwald
## Sample TCL based schedulers.
##
## Revision 1.1  1998/02/12 01:55:31  klauser
## CU-scheduler implementation revised for better computational complexity;
## easier portable to new scheduler hosts: all changes are localized in one
## file now;
##
#############################################################################

set timing 0

if {$timing} {
  puts "scheduler started  @ [exec date]"
}
set sleeptime 0

##
## get PBS status overview
##
set t [time {
  set servstat [pbsstatserv]
  set s_name [lindex $servstat 0]
  set s_state [getattrlval "server_state" $servstat]
  set s_cnt [getattrlval "state_count" $servstat]
  set s_nrunning [getcntval "Running" $s_cnt]
  set s_nqueued [getcntval "Queued" $s_cnt]
}]
set sleeptime [expr $sleeptime + [lindex $t 0]]
if {$timing} {
  puts "serverstat: [expr [lindex $t 0] / 1000] ms"
}
##
## fasttrack: if there are no queued jobs, we're done
##
if {$s_nqueued == 0} {
  if {$timing} {
    puts "scheduler finished @ [exec date]"
  }
  return 0
}

##
## get information on all queues
##
set t [time {
  kill_cpu_vars
  get_all_cpu_info
  erode_cpus
}]
set sleeptime [expr $sleeptime + [lindex $t 0]]
if {$timing} {
  puts "get_cpu_info: [expr [lindex $t 0] / 1000] ms"
}
##
## fasttrack: if there is no idle cpu, we're done
##
#if { [llength $all_cpus] == 0 } {
#  return
#}

##
## get information on all jobs
##
set t [time {
  kill_job_vars
  get_all_job_info
  erode_jobs
}]
set sleeptime [expr $sleeptime + [lindex $t 0]]
if {$timing} {
  puts "get_job_info: [expr [lindex $t 0] / 1000] ms"
}

#set debug2 1
##
## cleanup queued job debris
##
set t [time {
  collect_queued_jobs
}]
set sleeptime [expr $sleeptime + [lindex $t 0]]
if {$timing} {
  puts "cleanup_debris: [expr [lindex $t 0] / 1000] ms"
}

##
## The general idea of this scheduler is to enforce user job limits (number
## of running jobs) when necessary in high demand situations, but be more
## flexible in phases of low demand. This allows users with low job limits
## to borrow CPUs while they are not needed by anybody else, a situation that
## regularly occurs in our environment. Borrowed CPUs have to be given back
## when demand increases. This is done by rerunning (qrerun) the jobs. Rerun
## allows the running job to prepare for termination by first sending it an
## interceptable SIGTERM to e.g. cleanup or create a checkpoint.
##
## We apply a 3-level approach to scheduling:
## (1) constraint scheduling: 
##     empty CPUs to users within job limits
## (2) correction scheduling: 
##     users below job limit steal CPUs from users above job limit
## (3) auxiliary scheduling: 
##     empty CPUs to users above their job limit
##
set t [time {
  ##
  ## (1) constraint scheduling
  ##     try schedule while there are idle CPUs and queued jobs
  ##
  if {$debug2} {
    puts "scheduler: entering constraint scheduling phase"
    puts "sched_cnstr: $jobs_in_state(Q) jobs to schedule"
    puts "sched_cnstr: [llength $all_cpus] cpus to schedule"
  }
  set aux_cpus ""
  while { [llength $all_cpus] > 0 && $jobs_in_state(Q) > 0 } {
    ## cpu list is sorted in order of preference - use first cpu
    set cpu [lindex $all_cpus 0]
  
    ## we've got an idle cpu; schedule next job that can run on it
    set job [find_most_deserving_job $cpu]
    if { $job != 0 } {
      ## run this job on current cpu
      run_job_on_host $job $cpu
      if {$debug2} {
	puts "sched_cnstr: job $job scheduled on host $cpu"
      }

      ## check limit for this user after scheduling current job 
      check_user_limit $juser($job) "nosort"

      ## check limits for cpu after scheduling current job
      if { [is_cpu_busy $cpu] } {
	## remove cpu from scheduling pool
	set all_cpus [lreplace $all_cpus 0 0]
	if {$debug2} {
	  puts "sched_cnstr: host $cpu busy now"
	}
      }
    } else {
      ## no job found for this cpu - remove cpu from this scheduling pool
      ## and move it to auxiliary pool
      set all_cpus [lreplace $all_cpus 0 0]
      set aux_cpus [concat $aux_cpus $cpu]
      if {$debug2} {
	puts "sched_cnstr: no schedulable job for host $cpu"
      }
    }
  }
  ##
  ## (2) correction scheduling
  ##     try schedule while there are still jobs within user limits
  ##
  if {$debug2} {
    puts "scheduler: entering correction scheduling phase"
    puts "sched_corr: $jobs_aux_running auxiliary running jobs"
    puts "sched_corr: $jobs_in_state(Q) queued jobs to schedule"
  }
  while { $jobs_aux_running > 0 && $jobs_in_state(Q) > 0 } {
    ## we don't have a cpu yet
    set cpu ""
  
    ## get the best job
    set job [find_most_deserving_job $cpu]
    set user $juser($job)
    if { $job != 0 } {
      ## find a victim job and cpu for current job
      ## (juser must have cpu access permissions
      ##  and the cpu must still be enabled)
      set victim_job [find_least_deserving_aux_job $user]
      if { $victim_job != 0 } {
        set cpu $jqueue($victim_job)
        ## remove victim
        rerun_victim_job $victim_job $cpu
        ## run current job on this cpu
        run_job_on_host $job $cpu
        if {$debug2} {
	  puts "sched_corr: job $victim_job removed from host $cpu"
	  puts "sched_corr: job $job scheduled on host $cpu"
        }

        ## check limit for this user after scheduling current job 
        check_user_limit $user "nosort"
      } else {
        ## no victim_job found for this job 
        if {$debug2} {
	  puts "sched_corr: no victim job for job $job"
        }
        ## update state - remove job from list and count
        set idx [lsearch -exact $jobs_queued_by_user($user) $job]
        ## remove this job from queued list
        set jobs_queued_by_user($user) \
          [lreplace $jobs_queued_by_user($user) $idx $idx]
        incr jobs_in_state(Q) -1
      }
    } else {
      ## this is inconsistent with while condition - why are we here?
      puts "ERROR: sched_corr: inconsistent scheduler state"
      break
    }
  }
  ##
  ## (3) auxiliary scheduling
  ##     try schedule while there are still bored CPUs
  ##
  if {$debug2} {
    puts "scheduler: entering auxiliary scheduling phase"
  }
  ## restore auxiliary state back to normal state
  set all_cpus [concat $aux_cpus $all_cpus]
  restore_jobs_aux_queued
  if {$debug2} {
    puts "sched_aux: $jobs_in_state(Q) auxiliary jobs to schedule"
    puts "sched_aux: [llength $all_cpus] auxiliary cpus to schedule"
  }
  while { [llength $all_cpus] > 0 && $jobs_in_state(Q) > 0 } {
    ## cpu list is sorted in order of preference - use first cpu
    set cpu [lindex $all_cpus 0]
  
    ## we've got an idle cpu; schedule next job that can run on it
    set job [find_most_deserving_job $cpu]
    if { $job != 0 } {
      ## run this job on current cpu
      run_job_on_host $job $cpu
      if {$debug2} {
	puts "sched_aux: job $job scheduled on host $cpu"
      }

      ## check limits for cpu after scheduling current job
      if { [is_cpu_busy $cpu] } {
	## remove cpu from scheduling pool
	set all_cpus [lreplace $all_cpus 0 0]
	if {$debug2} {
	  puts "sched_aux: host $cpu busy now"
	}
      }
    } else {
      ## no job found for this cpu - remove cpu from this scheduling pool
      set all_cpus [lreplace $all_cpus 0 0]
      if {$debug2} {
	puts "sched_aux: no schedulable job for host $cpu"
      }
    }
  }
}]
set sleeptime [expr $sleeptime + [lindex $t 0]]
if {$timing} {
  puts "schedule_loop: [expr [lindex $t 0] / 1000] ms"
}

if {$timing} {
  puts "scheduler finished @ [exec date]"
}

#set debug2 0
##
## once we got here, we have consumed a lot of cpu cycles already;
## lets make sure that we shut up for a while before doing this again;
## this amortizes the scheduling effort over multiple scheduled jobs, 
## since the server coalesces successive scheduler trigger events into a 
## single scheduler activation.
##

##set sleeptime [expr $sleeptime / 1000000]
##if {$timing} {
##  puts "scheduler sleeping $sleeptime seconds"
##}
##sleep $sleeptime

