############################################################################
##                           PBS CU-scheduler
##
## CU-sched-common.tcl: common scheduling procedures
##
## (c) 1998  Dirk Grunwald and Artur Klauser, 
##           University of Colorado at Boulder,
##           [grunwald,klauser]@cs.colorado.edu
##
## $Id: CU-sched-common.tcl,v 2.0 1998/10/14 23:19:34 hender Exp $
## $Log: CU-sched-common.tcl,v $
## Revision 2.0  1998/10/14 23:19:34  hender
## PBS Release 2.0
##
## Revision 1.1  1998/08/06 21:20:12  hender
## PR #1234 - Dec Alpha mods and scheduler changes from U Colorado
##
## Revision 1.1  1998/08/05 22:20:57  grunwald
## Sample TCL based schedulers.
##
## Revision 1.3  1998/02/16 19:25:13  klauser
## fixed a few typos and qinfo problems
##
## Revision 1.2  1998/02/14 02:04:16  klauser
## fixed couple of typos
##
## Revision 1.1  1998/02/12 01:55:32  klauser
## CU-scheduler implementation revised for better computational complexity;
## easier portable to new scheduler hosts: all changes are localized in one
## file now;
##
#############################################################################

#############################################################################
## Parse PBS server data structure
#############################################################################
proc getattrlval {item sinfo} {
  global debug
  return [getval $item [lindex [lrange $sinfo 1 end] 0]]
}

#############################################################################
## Find value for key in {{key value} ...} list
#############################################################################
proc getval {item sinfo} {
  foreach pair $sinfo {
    set name [lindex $pair 0]
    set attrl [lindex $pair 1]
    if { $name == $item } {
      return $attrl
    }
  }
  return "";
}
#############################################################################
## Procedure for extracting various counts from the state count string
#############################################################################
proc getcntval {item cntinfo} {
  set cntstring [lindex $cntinfo [lsearch $cntinfo [set item]:* ]]
  return [lindex [split $cntstring ":"] 1]
}

#############################################################################
## Check if queue is up and running
#############################################################################
proc check_queue_status {name} {
  global qenabled
  global qtype
  global debug
  
  ##
  ## check for special queue names and return..
  ##
  foreach special $special_queue_names {
    if ( $name == $special ) {
      return;
    }
  }
      
  if { $qtype($name) != "Execution" } {
    if { $debug } {
      puts "Queue $name is not an execution queue"
    }
    return
  }
  ##
  ## Check that the system is up..
  ##
  set rmcon [openrm $name]
  addreq $rmcon "loadave"
  set status [getreq $rmcon]
  closerm $rmcon
  if { $status == "" } {
    puts "I think that $name is down..."
    if { $qenabled($name) } {
      disable_queue $name
    }
  } else {
    ##
    ## It's up - if we think it's down, let's restart it..
    ##
    puts "Looks like $name is up.."
#    if { ! $qenabled($name) } {
#      enable_queue $name
#    }
  }
}

#############################################################################
## Scan through all the jobs, and accumulate statistics on how many
## jobs each user has.
#############################################################################
proc get_all_job_info {} {
  global debug

  global jobs_in_state

  global jstate
  global jqueue
  global juser
  global jgroup
  global jrerunable

  global jobs_by_user
  global jobs_running_by_user
  global jobs_queued_by_user

  global jobs_by_group
  global jobs_running_by_group
  global jobs_queued_by_group

  global jobs_debris

  global special_queue_names

  global user_max_jobs
  global user_max_jobs_default

  kill_job_vars

  set jinfo [pbsstatjob]
  set all_jobs [llength $jinfo]

  puts "Found $all_jobs jobs"

  foreach jstruct $jinfo {
    set name [lindex $jstruct 0]
    if { $name != "" } {
      set user [getattrlval "euser" $jstruct]
      set state [getattrlval "job_state" $jstruct]

      set queue [getattrlval "queue" $jstruct]
#g#      set group [getattrlval "egroup" $jstruct]
      set rerunable [getattrlval "Rerunable" $jstruct]

      set jstate($name) $state
      set jqueue($name) $queue
      set juser($name) $user
#g#      set jgroup($name) $group
      if {$rerunable == "True"} {
        set jrerunable($name) 1
      } else {
        set jrerunable($name) 0
      }

      if { ! [ info exists jobs_by_user($user) ]} {
        set jobs_by_user($user) 0
        set jobs_running_by_user($user) ""
        set jobs_queued_by_user($user) ""
      }

      incr jobs_in_state($state) 1

#g#      if { ! [ info exists jobs_by_group($group) ]} {
#g#        set jobs_by_group($group) 0
#g#        set jobs_running_by_group($group) 0
#g#        set jobs_queued_by_group($group) 0
#g#      }

      incr jobs_by_user($user) 1
#g#      incr jobs_by_group($group) 1

      if { $state == "R" } {
        set jobs_running_by_user($user) \
          [concat $jobs_running_by_user($user) $name]
#g#        incr jobs_running_by_group($group) 1
      }

      if { $state == "Q" } {
        set jobs_queued_by_user($user) \
          [concat $jobs_queued_by_user($user) $name]
#g#        incr jobs_queued_by_group($group) 1
        ## queued jobs should always be in a special queue
	if { [lsearch -exact $special_queue_names $queue] == -1 } {
          set jobs_debris [concat $jobs_debris $name]
        }
      }
      ##
      ## define variables needed in rest of scheduler
      ##
      if { ! [info exists user_max_jobs($user)] } {
	## use default job limit
	set user_max_jobs($user) $user_max_jobs_default
      }
    }
  }
}

#############################################################################
## Find the most deserving job that can run on this cpu
#############################################################################
proc find_most_deserving_job {cpu} {
  global jobs_in_state

  global jstate
  global jqueue
  global juser
  global jgroup

  global jobs_by_user
  global jobs_running_by_user
  global jobs_queued_by_user

  global jobs_by_group
  global jobs_running_by_group
  global jobs_queued_by_group
  
  global user_max_jobs

  ## Find the person with the smallest number of running jobs
  set min_jobs 100000
  set min_person nobody
  foreach person [array names jobs_queued_by_user ] {
    ## A deserving person must have an account on the host
    if {   ($cpu == "" || [has_account $person $cpu])
        && [llength $jobs_queued_by_user($person)] > 0 } {
      if { $min_jobs > [llength $jobs_running_by_user($person)] } {
        set min_jobs [llength $jobs_running_by_user($person)]
        set min_person $person
      }
    }
  }

  if { $min_person == "nobody" } {
    ## No job found
    return 0
  } else {
    ## jobs are sorted by age - take oldest job
    set job [lindex $jobs_queued_by_user($min_person) 0]
    if {$job == ""} {
      puts "ERROR: find_most_deserving_job: empty job"
      return 0
    } else {
      return $job
    }
  }
}

#############################################################################
## Find the least deserving auxiliary job running, such that the
## specified user could run a job on this host instead
#############################################################################
proc find_least_deserving_aux_job {user} {
  global jobs_in_state

  global jstate
  global jqueue
  global juser
  global jgroup

  global jobs_by_user
  global jobs_running_by_user
  global jobs_aux_running_by_user
  global jobs_aux_running_l
  global jobs_aux_running

  global qenabled
  global user_max_jobs

  ## we use a cached sorted list of least deserving aux jobs
  ## if the list is not yet produced we have to make it first
  if { ! [info exists jobs_aux_running_l]} {
    make_jobs_aux_running_list
  }
  ## go through list and find first job which is running on a host
  ## that deserving user could also use
  set idx 0
  foreach job $jobs_aux_running_l {
    set cpu $jqueue($job)
    ## the cpu must still be enabled and
    ## the deserving person must have an account on the host
    if {   [info exists qenabled($cpu)] && $qenabled($cpu)
        && [has_account $user $cpu]} {
      ## remove job from aux_running list
      set jobs_aux_running_l [lreplace $jobs_aux_running_l $idx $idx]
      incr jobs_aux_running -1
      return $job
    }
    incr idx 1
  }
  return 0
}

#############################################################################
## Create a list of all aux jobs, sorted by how undeserving they are
#############################################################################
proc make_jobs_aux_running_list {} {
  global debug
  global debug2

  global jobs_in_state

  global jstate
  global jqueue
  global juser
  global jgroup

  global jobs_by_user
  global jobs_running_by_user
  global jobs_aux_running_by_user
  global jobs_aux_running_l
  global jobs_aux_running

  set jobs_aux_running_l ""
  while {[llength $jobs_aux_running_l] < $jobs_aux_running} {
    ## Find the person with the largest number of auxiliary running jobs
    set max_jobs 0
    set max_user nobody
    foreach user [array names jobs_aux_running_by_user ] {
      if {$jobs_aux_running_by_user($user) > $max_jobs} {
	set max_jobs $jobs_aux_running_by_user($user)
	set max_user $user
      }
    }

    if { $max_jobs == 0 } {
      puts "ERROR: make_jobs_aux_running_l: inconsistent data structures"
      return
    }
    ## Take that person's most recently started job
    set job [lindex $jobs_running_by_user($max_user) 0]
    if {$job == ""} {
      puts "ERROR: make_jobs_aux_running_l: empty job"
      return
    }
    if {$debug2} {
      puts "make_jobs_aux_running_list: added job $job user $max_user"
    }
    ## update data structures
    set jobs_running_by_user($max_user) \
      [lreplace $jobs_running_by_user($max_user) 0 0]
    set jobs_aux_running_l [concat $jobs_aux_running_l $job]
    ## note: jobs_aux_running_by_user gets destroyed by this procedure
    incr jobs_aux_running_by_user($max_user) -1
  }
}

#############################################################################
## Cleanup queued jobs on execution hosts and put them in central pool
#############################################################################
## qrerun leaves queued jobs on execution queues which must be moved
## back to the main scheduling pool in queue funnel
proc collect_queued_jobs {} {
  global debug
  global debug2

  global jqueue
  global jobs_debris
  global special_queue_names

  foreach job $jobs_debris {
    pbsmovejob $job "funnel"
    if {$debug2} {
      puts "moved queued job $job back to queue funnel"
    }
    set jqueue($job) funnel
  }
}

#############################################################################
## Get current status of all known queues
#############################################################################
proc get_all_cpu_info {} {
  global debug
  global debug2
  global qnames qcnt qtype nqueued max_queuable nrunning max_running
  global special_queue_names
  global qenabled

  global cpu_max_jobs
  global cpu_cur_jobs
  global cpu_cur_load
  global cpu_physmem
  global cpu_platform
  global cpu_hasvalid

  set server_cpus ""
  set qinfo [pbsstatque]

  foreach qstruct $qinfo {
    set name [lindex $qstruct 0]
    if { $name != "" } {
      set qnames($name) $name
      set qcnt($name) [getattrlval "state_count" $qstruct]
      set qtype($name) [getattrlval "queue_type" $qstruct]
      set nqueued($name) [getcntval "Queued" $qcnt($name)]
      set max_queuable($name) [getattrlval "max_queuable" $qstruct]
      if { [lsearch -exact $special_queue_names $name] == -1 } {
        set server_cpus [concat $server_cpus $name]
      }

      if { [getattrlval "enabled" $qstruct] == "True"} {
        set qenabled($name) 1
      } else {
        set qenabled($name) 0
      }

      if { $qtype($name) == "Execution" } {
        set nrunning($name) [getcntval "Running" $qcnt($name)]
        set max_running($name) [getattrlval "max_running" $qstruct]
      }
    }

    if { $debug2 } {
      puts "Found queue name $name of type $qtype($name)"
      puts "qcnt = $qcnt($name)"
      puts "qtype = $qtype($name)"
      puts "nqueued = $nqueued($name)"
      puts "queuable = $max_queuable($name)"
    }
  }

  ##
  ## Now, probe the CPU's
  ##
  foreach cpu $server_cpus {
    if [ info exists qenabled($cpu) ] {
      if { $qenabled($cpu) } {
        ##
        ## So some simple accounting changes..
        ##
        set cpu_cur_jobs($cpu) $nrunning($cpu)
        set cpu_cur_load($cpu) 0
        set cpu_max_jobs($cpu) 0

        set cpu_physmem($cpu) 0
        set cpu_hasvalid($cpu) "no"
        set cpu_platform($cpu) "unknown"

        set rmcon [openrm $cpu]
        addreq $rmcon "loadave"
        set loadave [getreq $rmcon]
        if { $loadave == "" } {
          puts "I think that $cpu is down (reply is *$loadave*)..."
          disable_queue $cpu
          closerm $rmcon
          puts "Disabled.."
          continue
        }
        addreq $rmcon "ncpus"
        set ncpu [getreq $rmcon]
        addreq $rmcon "physmem"
        set physmem [getreq $rmcon]
##
## Hack for new PBS standard for physmem..
##
	regsub kb $physmem "" physmem
        addreq $rmcon "platform"
##
## End of new PBS hack
##
        set platform [getreq $rmcon]
        addreq $rmcon valid_user\[user=root\]
        set hasvalid [getreq $rmcon]
        if { $hasvalid != "yes" } {
          set hasvalid "no"
        }
        ##
        ## Handle corner case of older OS
        ##
        if { $ncpu < 0 } { set ncpu 1 }
        closerm $rmcon

        set cpu_cur_load($cpu) $loadave
        set cpu_physmem($cpu) $physmem
        set cpu_platform($cpu) $platform
        set cpu_hasvalid($cpu) $hasvalid
        ##
        ## If we forgot to set the maximum number of running
        ## jobs, set it via the number of CPU's
        ##
        if { $max_running($cpu) == 0 } {
          set cpu_max_jobs($cpu) $ncpu
        } else {
          set cpu_max_jobs($cpu) $max_running($cpu)
        }
      }
    } else {
      puts "You need to create a queue for $cpu"
      set qenabled($cpu) 0
    }
  }
}

#############################################################################
## Print information overview for each queue
#############################################################################
proc print_cpu_info {cpulist} {
  global debug
  global debug2
  global qnames qcnt qtype nqueued max_queuable nrunning max_running
  global qenabled

  global cpu_max_jobs
  global cpu_cur_jobs
  global cpu_cur_load
  global cpu_physmem
  global cpu_platform
  global cpu_hasvalid

  puts [format "%-10s %6s %6s %6s %6s %4s %10s" "CPU" "MAX-J" "CUR-J" "LOAD" "MEM" "V?" "PLATFORM"]
  foreach cpu $cpulist {
    if { $qenabled($cpu) } {
      set physmem [expr ($cpu_physmem($cpu) + 1023) / 1024 ]
      puts [format "%-10s %6.1f %6.1f %6.1f %6d %4s %-35.35s" $cpu $cpu_max_jobs($cpu) $cpu_cur_jobs($cpu) $cpu_cur_load($cpu) $physmem $cpu_hasvalid($cpu) $cpu_platform($cpu)]
    } else {
      puts "$cpu\t\t***disabled***";
    }
  }
}


#############################################################################
## Get cluster names
#############################################################################

proc get_all_clusters {} {
  global debug
  global debug2

  global cpu_pool

  set these_clusters ""

  foreach cluster $cpu_pool {
    set these_clusters [concat $these_clusters [getval "cluster" $cluster]]
  }

  return $these_clusters
}

#############################################################################
## Get host names from a cluster specification
#############################################################################

proc get_hosts_from_cluster_spec {cluster_names} {

  global debug
  global debug2

  global cpu_pool

  ##
  ## hosts is the return variable
  ##

  foreach cluster_name $cluster_names {
    foreach cluster $cpu_pool {
      set flags [getval "flags" $cluster]
      set cname [getval "cluster" $cluster]
      if { $cname == $cluster_name } {
        if { [lsearch -exact $flags "noschedule"] != -1 } {
          ## machines in noschedule clusters are dropped immediately
          if { $debug2 } {
            puts "found cluster $cname ($flags), ignored"
          }
          continue
        }
        set hosts [getval "machines" $cluster]
        return [lsort $hosts]
      }
    }
  }
  return ""
}


#############################################################################
## Find all schedulable CPUs
#############################################################################
proc erode_cpus {} {
  global debug
  global debug2
  global qnames qcnt qtype nqueued max_queuable nrunning max_running
  global qenabled

  global cpu_max_jobs
  global cpu_cur_jobs
  global cpu_cur_load

  global cpu_pool
  global cpu_flags
  global cpu_cluster
  global all_cpus

  ##
  ## Beginning with the cpu_pool, we eliminate all cpus which can't handle
  ## any more jobs, and keep the rest in all_cpus, sorted by priority
  ##
  set all_cpus ""

  ## step through cpu clusters
  foreach cluster $cpu_pool {
    set flags [getval "flags" $cluster]
    set cname [getval "cluster" $cluster]
    if { [lsearch -exact $flags "noschedule"] != -1 } {
      ## machines in noschedule clusters are dropped immediately
      if { $debug2 } {
        puts "found cluster $cname ($flags), ignored"
      }
      continue
    }
    if { [lsearch -exact $flags "onlyidle"] == -1 } {
      ## schedule on these hosts independent of load
      set hosts [getval "machines" $cluster]
      if { $debug2 } {
        puts "found cluster $cname ($flags)"
      }
    } else {
      ## schedule on least loaded host first - sorting ensures right prio
      set hosts [lsort -command hostsort [getval "machines" $cluster]]
      if { $debug2 } {
        puts "found cluster $cname ($flags), sorted by load"
      }
    }
    ## step through hosts
    foreach host $hosts {
      if { ! [info exists qenabled($host)]} {
	## this can happen if the server does ommits status information
        ## from unreachable hosts
        set qenabled($host) 0
      }
      if { $qenabled($host) } {
        if { [is_cpu_busy $host] } {
          if { $debug2 } {
            puts "  host $host already busy"
          }
        } else {
          set all_cpus [concat $all_cpus $host]
          if { $debug2 } {
            puts "  host $host schedulable"
          }
        }
      } else {
        if { $debug2 } {
          puts "  host $host is disabled"
        }
      }
    }
  }
}
  
#############################################################################
## sort hosts by increasing load
#############################################################################
proc hostsort {a b} {
  global cpu_cur_load
  global cpu_max_jobs
 
  set aload 1
  set bload 1
  if {   [ info exists cpu_cur_load($a) ] \
      && [ info exists cpu_max_jobs($a) ] } {
    if { $cpu_max_jobs($a) > 0 } {
      set aload [expr $cpu_cur_load($a) / $cpu_max_jobs($a)]
    }
  }
  if {   [ info exists cpu_cur_load($b) ] \
      && [ info exists cpu_max_jobs($b) ] } {
    if { $cpu_max_jobs($b) > 0 } {
      set bload [expr $cpu_cur_load($b) / $cpu_max_jobs($b)]
    }
  }
  set retval [expr $aload - $bload]
  if {$retval > 0} {
    return 1
  } elseif {$retval < 0} {
    return -1
  } else {
    return 0
  }
}

#############################################################################
## check if CPU is idle or busy
#############################################################################
proc is_cpu_busy {host} {
  global debug
  global debug2
  global cpu_flags
  global cpu_cur_jobs
  global cpu_cur_load
  global cpu_max_jobs

  if { [lsearch -exact $cpu_flags($host) "onlyidle"] == -1 } {
    ## schedule on this host independent of load;
    ## can take up to max_jobs of PBS work
    if { [ info exists cpu_cur_jobs($host) ] \
      && [ expr $cpu_cur_jobs($host) < $cpu_max_jobs($host) ] } \
    {
      return 0
    } else {
      return 1
    }
  } else {
    ## schedule only if host is lightly loaded,
    ## can take up to max_jobs of any work
    if { [ info exists cpu_cur_jobs($host) ] \
      && [ info exists cpu_cur_load($host) ] \
      && [ expr $cpu_cur_jobs($host) < $cpu_max_jobs($host) ] \
      && [ expr $cpu_cur_load($host) < $cpu_max_jobs($host) ] } \
    {
      return 0
    } else {
      return 1
    }
  }
}
      
#############################################################################
## find all schedulable jobs
#############################################################################
proc erode_jobs {} {
  global debug
  global debug2

  global jobs_queued_by_user
  global jobs_running_by_user

  global jobs_aux_running
  global jobs_aux_running_by_user

  global user_max_jobs

  ## eliminate all jobs by users which are at the job limit
  foreach user [array names jobs_queued_by_user] {
    check_user_limit $user "sort"
  }

  ## determine number of auxiliary jobs running
  set jobs_aux_running 0
  foreach user [array names jobs_running_by_user] {
    set aux_jobs \
      [expr [llength $jobs_running_by_user($user)] - $user_max_jobs($user)]
    if {$aux_jobs < 0} {
      set aux_jobs 0
    }
    set jobs_aux_running_by_user($user) $aux_jobs
    incr jobs_aux_running $aux_jobs
    set jobs_running_by_user($user) \
      [lsort -command jobsort -decreasing $jobs_running_by_user($user)]
    if {$debug2} {
      puts "$aux_jobs auxiliary running jobs for user $user"
    }
  }
}

#############################################################################
## check if user is at job limit
#############################################################################
proc check_user_limit {user sorting} {
  global debug
  global debug2
  global jobs_in_state

  global jstate
  global jqueue
  global juser
  global jgroup

  global jobs_by_user
  global jobs_running_by_user
  global jobs_queued_by_user
  global jobs_aux_queued_by_user

  global jobs_by_group
  global jobs_running_by_group
  global jobs_queued_by_group
    
  global user_max_jobs

  set nqjobs [llength $jobs_queued_by_user($user)]
  if { $nqjobs == 0 } {
    unset jobs_queued_by_user($user)
  } else {
    if {$sorting == "sort"} {
      set jobs_queued_by_user($user) \
	[lsort -command jobsort $jobs_queued_by_user($user)]
      if {$debug2} {
	puts "user $user $nqjobs queued jobs sorted"
      }
    }
    if { [llength $jobs_running_by_user($user)] >= $user_max_jobs($user)} {
      ## user is at job limit - eliminate their jobs from scheduling pool
      incr jobs_in_state(Q) -$nqjobs
      incr jobs_by_user($user) -$nqjobs
      set jobs_aux_queued_by_user($user) $jobs_queued_by_user($user)
      unset jobs_queued_by_user($user)
      if {$debug2} {
        puts "user $user at job limit of $user_max_jobs($user)"
      }
    } else {
      ## nada
    }
  }
}

#############################################################################
## sort jobs by jobid (= time resident in PBS)
#############################################################################
proc jobsort {a b} {
  set anum [lindex [split $a {.}] 0]
  set bnum [lindex [split $b {.}] 0]
  return [expr $anum - $bnum]
}

#############################################################################
## move auxiliary queued jobs back to normal queue
#############################################################################
proc restore_jobs_aux_queued {} {
  global debug
  global debug2

  global jrerunable

  global jobs_in_state

  global jobs_by_user
  global jobs_running_by_user
  global jobs_queued_by_user
  global jobs_aux_queued_by_user

  global jobs_by_group
  global jobs_running_by_group
  global jobs_queued_by_group

  foreach user [array names jobs_aux_queued_by_user ] {
    set jobs_queued_by_user($user) ""
    foreach job $jobs_aux_queued_by_user($user) {
      if {$jrerunable($job)} {
	incr jobs_in_state(Q) 1
	incr jobs_by_user($user) 1
	set jobs_queued_by_user($user) \
	  [concat $jobs_queued_by_user($user) $job]
      }
    }
    if {$debug2} {
      puts "[llength $jobs_queued_by_user($user)] of [llength $jobs_aux_queued_by_user($user)] auxiliary jobs queued for user $user"
    }
    unset jobs_aux_queued_by_user($user)
    if {$jobs_queued_by_user($user) == ""} {
      unset jobs_queued_by_user($user)
    }
  }
}
    
#############################################################################
## remove job from host
## update all job and cpu statistics accordingly
#############################################################################
proc rerun_victim_job {job_to_schedule host} {
  global juser
  global jobs_by_user
  global jobs_running_by_user
  global jobs_queued_by_user
  global jobs_in_state

  global cpu_max_jobs
  global cpu_cur_jobs
  global cpu_cur_load
  global jstate

  ## move job to the queue for that host, and then run the job
  set user $juser($job_to_schedule)
  puts "Rerun $job_to_schedule from $host for $user"
  pbsrerun $job_to_schedule
  pbsmovejob $job_to_schedule "funnel"

  ## Now, update various statistics used to determine "fair" scheduling
  ## Note: we don't put this job back on any queued list, so we don't try
  ## to restart it again in a laster phase of this scheduling cycle
  incr jobs_in_state(R) -1
  incr cpu_cur_jobs($host) -1
  set cpu_cur_load($host) [expr $cpu_cur_load($host) - 1]
  set jstate($job_to_schedule) Q
}

#############################################################################
## put job on host
## update all job and cpu statistics accordingly
#############################################################################
proc run_job_on_host {job_to_schedule host} {
  global juser
  global jobs_by_user
  global jobs_running_by_user
  global jobs_queued_by_user
  global jobs_in_state

  global cpu_max_jobs
  global cpu_cur_jobs
  global cpu_cur_load
  global jstate

  ## move job to the queue for that host, and then run the job
  set user $juser($job_to_schedule)
  puts "Run $job_to_schedule on $host for $user"
  pbsmovejob $job_to_schedule $host
  pbsrunjob $job_to_schedule $host

  ## Now, update various statistics used to determine "fair" scheduling
  set idx [lsearch -exact $jobs_queued_by_user($user) $job_to_schedule]
  ## remove this job from queued list
  set jobs_queued_by_user($user) \
    [lreplace $jobs_queued_by_user($user) $idx $idx]
  ## add this job to running list
  set jobs_running_by_user($user) \
    [concat $jobs_running_by_user($user) $job_to_schedule]
  incr jobs_in_state(Q) -1
  incr jobs_in_state(R) 1
  incr cpu_cur_jobs($host) 1
  set cpu_cur_load($host) [expr $cpu_cur_load($host) + 1]
  set jstate($job_to_schedule) R
}

#############################################################################
## check if user has a valid account on host
## we use a user-host cache to avoid repeating queries to mom;
## if users are added/removed from hosts, the scheduler must be restared 
## to clear this cache! (happens very rarely for us)
#############################################################################
proc has_account {user host} {
  global uh_cache

  if { ! [info exists uh_cache($user@$host)]} {
    set c [ openrm $host ]
    addreq $c valid_user\[user=$user\]
    set v [getreq $c]
    closerm $c
    if { $v == "yes" } {
      set uh_cache($user@$host) 1
    } else {
      set uh_cache($user@$host) 0
    }
  }
  return $uh_cache($user@$host)
}

#############################################################################
## disable a queue
#############################################################################
proc disable_queue {name} {
  global debug
  global PBS qenabled
  global disable_queues_if_no_reply

  if { $debug } {
    puts "Disable  queue $name"
  }
  if { $disable_queues_if_no_reply } {
      pbsqdisable $name
  }
  set qenabled($name) 0
}

#############################################################################
## enable a queue
#############################################################################
proc enable_queue {name} {
  global debug
  global PBS qenabled
  if { $debug } {
    puts "Enable queue $name"
  }
  pbsqenable $name
  set qenabled($name) 1
}

#############################################################################
## unset all cpu variables to avoid garbage..
#############################################################################
proc kill_cpu_vars {} {
  global qnames qcnt qtype nqueued max_queuable nrunning max_running
  global qenabled

  global cpu_max_jobs
  global cpu_cur_jobs
  global cpu_cur_load
  global cpu_physmem
  global cpu_platform
  global cpu_hasvalid

  if [ info exists qnames ] { unset qnames }
  if [ info exists qcnt ] { unset qcnt }
  if [ info exists qtype ] { unset qtype }
  if [ info exists nqueued ] { unset nqueued }
  if [ info exists max_queuable ] { unset max_queuable }
  if [ info exists nrunning ] { unset nrunning }
  if [ info exists max_running ] { unset max_running }

  if [ info exists qenabled ] { unset qenabled }
  if [ info exists cpu_max_jobs ] { unset cpu_max_jobs }
  if [ info exists cpu_cur_jobs ] { unset cpu_cur_jobs }
  if [ info exists cpu_cur_load ] { unset cpu_cur_load }
  if [ info exists cpu_physmem ] { unset cpu_physmem }
  if [ info exists cpu_platform ] { unset cpu_platform }
  if [ info exists cpu_hasvalid ] { unset cpu_hasvalid }
}

#############################################################################
## unset all job variables to avoid garbage..
#############################################################################
proc kill_job_vars {} {
  global jobs_in_state
  global jobs_aux_running

  global jstate
  global jqueue
  global juser
  global jgroup
  global jrerunable

  global jobs_by_user
  global jobs_running_by_user
  global jobs_aux_running_by_user
  global jobs_aux_running_l
  global jobs_aux_running
  global jobs_queued_by_user
  global jobs_aux_queued_by_user

  global jobs_by_group
  global jobs_running_by_group
  global jobs_queued_by_group

  global jobs_debris

  if [ info exists jobs_in_state ] { unset jobs_in_state }
  if [ info exists jstate ] { unset jstate }
  if [ info exists jqueue ] { unset jqueue }
  if [ info exists juser ] { unset juser }
  if [ info exists jgroup ] { unset jgroup }
  if [ info exists jrerunable ] { unset jrerunable }

  if [ info exists jobs_by_user ] { unset jobs_by_user }
  if [ info exists jobs_running_by_user ] { unset jobs_running_by_user }
  if [ info exists jobs_aux_running_by_user ] { unset jobs_aux_running_by_user }
  if [ info exists jobs_aux_running_l ] { unset jobs_aux_running_l }
  if [ info exists jobs_aux_running ] { unset jobs_aux_running }
  if [ info exists jobs_queued_by_user ] { unset jobs_queued_by_user }
  if [ info exists jobs_aux_queued_by_user ] { unset jobs_aux_queued_by_user }

  if [ info exists jobs_by_group ] { unset jobs_by_group }
  if [ info exists jobs_running_by_group ] { unset jobs_running_by_group }
  if [ info exists jobs_queued_by_group ] { unset jobs_queued_by_group }

  set jobs_in_state(E) 0
  set jobs_in_state(H) 0
  set jobs_in_state(Q) 0
  set jobs_in_state(R) 0
  set jobs_in_state(S) 0
  set jobs_in_state(T) 0
  set jobs_in_state(W) 0

  set jobs_debris ""
}
